/*
   MicroDexed

   MicroDexed is a port of the Dexed sound engine
   (https://github.com/asb2m10/dexed) for the Teensy-3.5/3.6 with audio shield.
   Dexed ist heavily based on https://github.com/google/music-synthesizer-for-android

   (c)2018,2019 H. Wirtz <wirtz@parasitstudio.de>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
*/

#include <limits.h>
#include "config.h"
#include <Audio.h>
#include <Wire.h>
#include <SPI.h>
#include <SD.h>
#include <MIDI.h>
#include <EEPROM.h>
#include "EEPROMAnything.h"
#include "midi_devices.hpp"
#include "dexed.h"
#include "dexed_sysex.h"
#include "effect_modulated_delay.h"
#include "effect_stereo_mono.h"
#include "PluginFx.h"
#include "SoftenValue.hpp"
#include "UI.hpp"
#include "source_microdexed.h"

// Audio engines
AudioAnalyzePeak         peak1;
AudioSourceMicroDexed* MicroDexed[NUM_DEXED];
AudioEffectDelay         delay1;
AudioEffectModulatedDelay   modchorus;
AudioSynthWaveform       modulator;
AudioMixer4              delay_fb_mixer;
AudioMixer4              master_mixer_r;
AudioMixer4              master_mixer_l;
AudioAmplifier           volume_r;
AudioAmplifier           volume_l;
AudioAmplifier           modchorus_inverter;
#if defined(AUDIO_DEVICE_USB)
AudioOutputUSB           usb1;
#endif
AudioEffectStereoMono    stereomono1;

// Audio connections
AudioConnection          patchCord1(modulator, 0, modchorus, 1);
#if MOD_FILTER_OUTPUT != MOD_NO_FILTER_OUTPUT
AudioFilterBiquad        modchorus_filter;
AudioConnection          patchCord2(modchorus, modchorus_filter);
AudioConnection          patchCord3(modchorus_filter, 0, master_mixer_r, 3);
AudioConnection          patchCord4(modchorus_filter, modchorus_inverter);
AudioConnection          patchCord5(modchorus_inverter, 0, master_mixer_l, 3);
#else
AudioConnection          patchCord6(modchorus, 0, master_mixer_r, 3);
AudioConnection          patchCord7(modchorus, modchorus_inverter);
AudioConnection          patchCord8(modchorus_inverter, 0, master_mixer_l, 3);
#endif
#if defined(USE_REVERB)
AudioEffectFreeverbStereo freeverbs1;
AudioConnection          patchCord9(freeverbs1, 0, master_mixer_r, 1);
AudioConnection          patchCord10(freeverbs1, 1, master_mixer_l, 1);
#endif
AudioConnection          patchCord11(delay_fb_mixer, delay1);
AudioConnection          patchCord12(delay1, 0, delay_fb_mixer, 1);
AudioConnection          patchCord13(delay1, 0, master_mixer_r, 2);
AudioConnection          patchCord14(delay1, 0, master_mixer_l, 2);
AudioConnection          patchCord15(master_mixer_r, volume_r);
AudioConnection          patchCord16(master_mixer_l, volume_l);
AudioConnection          patchCord17(volume_r, 0, stereomono1, 0);
AudioConnection          patchCord18(volume_l, 0, stereomono1, 1);
#ifdef AUDIO_DEVICE_USB
AudioConnection          patchCord19(stereomono1, 0, usb1, 0);
AudioConnection          patchCord20(stereomono1, 1, usb1, 1);
#endif
#if defined(TEENSY_AUDIO_BOARD)
AudioOutputI2S           i2s1;
AudioConnection          patchCord21(stereomono1, 0, i2s1, 0);
AudioConnection          patchCord22(stereomono1, 1, i2s1, 1);
AudioControlSGTL5000     sgtl5000_1;
#elif defined (I2S_AUDIO_ONLY)
AudioOutputI2S           i2s1;
AudioConnection          patchCord23(stereomono1, 0, i2s1, 0);
AudioConnection          patchCord24(stereomono1, 1, i2s1, 1);
#elif defined(TGA_AUDIO_BOARD)
AudioOutputI2S           i2s1;
AudioConnection          patchCord25(stereomono1, 0, i2s1, 0);
AudioConnection          patchCord26(stereomono1, 1, i2s1, 1);
AudioControlWM8731master wm8731_1;
#elif defined(PT8211_AUDIO)
AudioOutputPT8211        pt8211_1;
AudioConnection          patchCord27(stereomono1, 0, pt8211_1, 0);
AudioConnection          patchCord28(stereomono1, 1, pt8211_1, 1);
#elif defined(TEENSY_DAC_SYMMETRIC)
AudioOutputAnalogStereo  dacOut;
AudioMixer4              invMixer;
AudioConnection          patchCord29(stereomono1, 0, dacOut  , 0);
AudioConnection          patchCord30(stereomono1, 1, invMixer, 0);
AudioConnection          patchCord31(invMixer, 0, dacOut  , 1);
#else
AudioOutputAnalogStereo  dacOut;
AudioConnection          patchCord32(stereomono1, 0, dacOut, 0);
AudioConnection          patchCord33(stereomono1, 1, dacOut, 1);
#endif
AudioConnection * dynamicConnections[NUM_DEXED * 7];
int nDynamic = 0;

void create_audio_connections(AudioSourceMicroDexed &d)
{
  dynamicConnections[nDynamic++] = new AudioConnection(d, peak1);
  dynamicConnections[nDynamic++] = new AudioConnection(d, 0, delay_fb_mixer, 0);
  dynamicConnections[nDynamic++] = new AudioConnection(d, 0, modchorus, 0);
  dynamicConnections[nDynamic++] = new AudioConnection(d, 0 , master_mixer_r, 0);
  dynamicConnections[nDynamic++] = new AudioConnection(d, 0 , master_mixer_l, 0);
#if defined(USE_REVERB)
  dynamicConnections[nDynamic++] = new AudioConnection(d, 0, freeverbs1, 0);
#endif
  for (int i = 0; i < nDynamic; i++) {
    if (dynamicConnections[i] == NULL)
      Serial.printf ("Bad connection %d\n", i);
  }
}

bool sd_card_available = false;
uint8_t max_loaded_banks = 0;
char bank_name[BANK_NAME_LEN];
char voice_name[VOICE_NAME_LEN];
char bank_names[MAX_BANKS][BANK_NAME_LEN];
char voice_names[MAX_VOICES][VOICE_NAME_LEN];
elapsedMillis autostore;
uint8_t midi_timing_counter = 0; // 24 per qarter
elapsedMillis midi_timing_timestep;
uint16_t midi_timing_quarter = 0;
elapsedMillis long_button_pressed;
SoftenValue <uint8_t> soften_volume;
SoftenValue <uint8_t> soften_filter_res[NUM_DEXED];
SoftenValue <uint8_t> soften_filter_cut[NUM_DEXED];
//elapsedMicros fill_audio_buffer;
elapsedMillis control_rate;
uint8_t active_voices = 0;
#ifdef SHOW_CPU_LOAD_MSEC
elapsedMillis cpu_mem_millis;
#endif
//uint32_t overload = 0;
uint32_t peak = 0;

config_t configuration;

bool eeprom_update_flag = false;

// Allocate the delay lines for chorus
int16_t delayline[MOD_DELAY_SAMPLE_BUFFER];

#ifdef ENABLE_LCD_UI
/***********************************************************************
   LCDMenuLib2
 ***********************************************************************/
extern LCDMenuLib2 LCDML;
extern uint8_t menu_state;
#endif

#ifdef DISPLAY_LCD_SPI
void change_disp_sd(bool disp)
{
  digitalWrite(SDCARD_CS_PIN, disp);
  digitalWrite(U8X8_CS_PIN, !disp);
}
#endif

void setup()
{
  //while (!Serial) ; // wait for Arduino Serial Monitor
  Serial.begin(SERIAL_SPEED);
#ifdef DISPLAY_LCD_SPI
  pinMode(SDCARD_CS_PIN, OUTPUT);
  pinMode(U8X8_CS_PIN, OUTPUT);
#endif

  delay(220); // necessary, because before this time no serial output is done :(

#ifdef ENABLE_LCD_UI
  setup_ui();
#else
  Serial.println(F("NO LCD DISPLAY ENABLED!"));
#endif

  Serial.println(F("MicroDexed based on https://github.com/asb2m10/dexed"));
  Serial.println(F("(c)2018,2019 H. Wirtz <wirtz@parasitstudio.de>"));
  Serial.println(F("https://codeberg.org/dcoredump/MicroDexed"));
  Serial.print(F("Version: "));
  Serial.println(VERSION);
  Serial.println(F("<setup start>"));

  initial_values_from_eeprom();

  for (uint8_t i = 0; i < NUM_DEXED; i++)
  {
    Serial.print(F("Creating MicroDexed engine "));
    Serial.println(i, DEC);
    MicroDexed[i] = new AudioSourceMicroDexed(SAMPLE_RATE);
    if (MicroDexed[i] == NULL) {
          Serial.print(F("Unable to create MicroDexed engine!"));
    }
    create_audio_connections(*MicroDexed[i]);
  }

  setup_midi_devices();

  // start audio card
  AudioNoInterrupts();
  AudioMemory(AUDIO_MEM);

#if defined(TEENSY_AUDIO_BOARD)
  sgtl5000_1.enable();
  sgtl5000_1.dacVolumeRamp();
  //sgtl5000_1.dacVolumeRampLinear();
  //sgtl5000_1.dacVolumeRampDisable();
  sgtl5000_1.unmuteHeadphone();
  sgtl5000_1.unmuteLineout();
  sgtl5000_1.autoVolumeDisable(); // turn off AGC
  sgtl5000_1.volume(0.5, 0.5); // Headphone volume
  sgtl5000_1.lineOutLevel(SGTL5000_LINEOUT_LEVEL);
  sgtl5000_1.audioPostProcessorEnable();
  sgtl5000_1.autoVolumeControl(1, 1, 1, 0.9, 0.01, 0.05);
  sgtl5000_1.autoVolumeEnable();
  sgtl5000_1.surroundSoundEnable();
  sgtl5000_1.surroundSound(7, 2); // Configures virtual surround width from 0 (mono) to 7 (widest). select may be set to 1 (disable), 2 (mono input) or 3 (stereo input).
  sgtl5000_1.enhanceBassEnable();
  sgtl5000_1.enhanceBass(1.0, 0.2, 1, 2); // Configures the bass enhancement by setting the levels of the original stereo signal and the bass-enhanced mono level which will be mixed together. The high-pass filter may be enabled (0) or bypassed (1).
  /* The cutoff frequency is specified as follows:
    value  frequency
    0      80Hz
    1     100Hz
    2     125Hz
    3     150Hz
    4     175Hz
    5     200Hz
    6     225Hz
  */
  //sgtl5000_1.eqBands(bass, mid_bass, midrange, mid_treble, treble);
  Serial.println(F("Teensy-Audio-Board enabled."));
#elif defined(TGA_AUDIO_BOARD)
  wm8731_1.enable();
  wm8731_1.volume(1.0);
  Serial.println(F("TGA board enabled."));
#elif defined(I2S_AUDIO_ONLY)
  Serial.println(F("I2S enabled."));
#elif defined(PT8211_AUDIO)
  Serial.println(F("PT8211 enabled."));
#elif defined(TEENSY_DAC_SYMMETRIC)
  invMixer.gain(0, -1.f);
  Serial.println(F("Internal DAC using symmetric outputs enabled."));
#else
  Serial.println(F("Internal DAC enabled."));
#endif

  // start SD card
#ifndef TEENSY4
  SPI.setMOSI(SDCARD_MOSI_PIN);
  SPI.setSCK(SDCARD_SCK_PIN);
#endif
#ifdef DISPLAY_LCD_SPI
  change_disp_sd(false);
#endif
  if (!SD.begin(SDCARD_CS_PIN))
  {
    Serial.println(F("SD card not accessable."));
    strcpy(bank_name, "Default");
    strcpy(voice_name, "Default");
  }
  else
  {
    Serial.println(F("SD card found."));
    sd_card_available = true;

    // read all bank names
    max_loaded_banks = get_bank_names();
    strip_extension(bank_names[configuration.bank], bank_name);

    // read all voice name for actual bank
    get_voice_names_from_bank(configuration.bank);
#ifdef DEBUG
    Serial.print(F("Bank ["));
    Serial.print(bank_names[configuration.bank]);
    Serial.print(F("/"));
    Serial.print(bank_name);
    Serial.println(F("]"));
    for (uint8_t n = 0; n < MAX_VOICES; n++)
    {
      if (n < 10)
        Serial.print(F(" "));
      Serial.print(F("   "));
      Serial.print(n, DEC);
      Serial.print(F("["));
      Serial.print(voice_names[n]);
      Serial.println(F("]"));
    }
#endif

    // load default SYSEX data
    load_sysex(configuration.bank, configuration.voice);
  }
#ifdef DISPLAY_LCD_SPI
  change_disp_sd(true);
#endif
  // Init effects
  memset(delayline, 0, sizeof(delayline));
  if (!modchorus.begin(delayline, MOD_DELAY_SAMPLE_BUFFER)) {
    Serial.println(F("AudioEffectModulatedDelay - begin failed"));
    while (1);
  }
#ifdef DEBUG
  Serial.print(F("MOD_DELAY_SAMPLE_BUFFER="));
  Serial.print(MOD_DELAY_SAMPLE_BUFFER, DEC);
  Serial.println(F(" samples"));
#endif
  master_mixer_r.gain(0, 1.0);
  master_mixer_l.gain(0, 1.0);

#ifdef USE_REVERB
  // INIT REVERB
  freeverbs1.roomsize(mapfloat(configuration.reverb_roomsize, REVERB_ROOMSIZE_MIN, REVERB_ROOMSIZE_MAX, 0.0, 1.0));
  freeverbs1.damping(mapfloat(configuration.reverb_damping, REVERB_DAMPING_MIN, REVERB_DAMPING_MAX, 0.0, 1.0));
  master_mixer_r.gain(1, mapfloat(configuration.reverb_level, REVERB_LEVEL_MIN, REVERB_LEVEL_MAX, 0.0, 1.0));
  master_mixer_l.gain(1, mapfloat(configuration.reverb_level, REVERB_LEVEL_MIN, REVERB_LEVEL_MAX, 0.0, 1.0));
#endif

  // INIT DELAY
  delay1.delay(0, mapfloat(configuration.delay_time, DELAY_TIME_MIN, DELAY_TIME_MAX, 0.0, float(DELAY_TIME_MAX)));
  // delay_fb_mixer is the feedback-adding mixer
  //delay_fb_mixer.gain(0, 1.0 - (mapfloat(configuration.delay_feedback, DELAY_FEEDBACK_MIN, DELAY_FEEDBACK_MAX, 0.0, 1.0))); // original signal
  delay_fb_mixer.gain(0, 1.0); // original signal
  delay_fb_mixer.gain(1, mapfloat(configuration.delay_feedback, DELAY_FEEDBACK_MIN, DELAY_FEEDBACK_MAX, 0.0, 1.0)); // amount of feedback
  master_mixer_r.gain(2, mapfloat(configuration.delay_level, DELAY_LEVEL_MIN, DELAY_LEVEL_MAX, 0.0, 1.0));
  master_mixer_l.gain(2, mapfloat(configuration.delay_level, DELAY_LEVEL_MIN, DELAY_LEVEL_MAX, 0.0, 1.0));

  // INIT CHORUS
  switch (configuration.chorus_waveform)
  {
    case 0:
      modulator.begin(WAVEFORM_TRIANGLE);
      break;
    case 1:
      modulator.begin(WAVEFORM_SINE);
      break;
    default:
      modulator.begin(WAVEFORM_TRIANGLE);
  }
  modulator.phase(0);
  modulator.amplitude(mapfloat(configuration.chorus_depth, CHORUS_DEPTH_MIN, CHORUS_DEPTH_MAX, 0.0, 1.0));
  modulator.offset(0.0);
#if MOD_FILTER_OUTPUT == MOD_BUTTERWORTH_FILTER_OUTPUT
  // Butterworth filter, 12 db/octave
  modchorus_filter.setLowpass(0, MOD_FILTER_CUTOFF_HZ, 0.707);
#elif MOD_FILTER_OUTPUT == MOD_LINKWITZ_RILEY_FILTER_OUTPUT
  // Linkwitz-Riley filter, 48 dB/octave
  modchorus_filter.setLowpass(0, MOD_FILTER_CUTOFF_HZ, 0.54);
  modchorus_filter.setLowpass(1, MOD_FILTER_CUTOFF_HZ, 1.3);
  modchorus_filter.setLowpass(2, MOD_FILTER_CUTOFF_HZ, 0.54);
  modchorus_filter.setLowpass(3, MOD_FILTER_CUTOFF_HZ, 1.3);
#endif
  if (configuration.mono == 0)
    modchorus_inverter.gain(-1.0); // stereo mode
  else
    modchorus_inverter.gain(1.0); // mono mode
  master_mixer_r.gain(3, mapfloat(configuration.chorus_level, CHORUS_LEVEL_MIN, CHORUS_LEVEL_MAX, 0.0, 1.0));
  master_mixer_l.gain(3, mapfloat(configuration.chorus_level, CHORUS_LEVEL_MIN, CHORUS_LEVEL_MAX, 0.0, 1.0));

  for (uint8_t i = 0; i < NUM_DEXED; i++)
  {
    soften_filter_res[i].init(configuration.filter_resonance);
    soften_filter_cut[i].init(configuration.filter_cutoff);
    MicroDexed[i]->fx.Gain =  mapfloat(configuration.loudness, LOUDNESS_MIN, LOUDNESS_MAX, 0.0, 1.0);
    MicroDexed[i]->fx.Reso = mapfloat(configuration.filter_resonance, FILTER_RESONANCE_MIN, FILTER_RESONANCE_MAX, 0.0, 1.0);
    MicroDexed[i]->fx.Cutoff = mapfloat(configuration.filter_cutoff, FILTER_CUTOFF_MIN, FILTER_CUTOFF_MAX, 0.0, 1.0);
  }

  // set initial volume and pan (read from EEPROM)
  set_volume(configuration.vol, configuration.pan, configuration.mono);
  soften_volume.init(configuration.vol);

#if defined (DEBUG) && defined (SHOW_CPU_LOAD_MSEC)
  // Initialize processor and memory measurements
  AudioProcessorUsageMaxReset();
  AudioMemoryUsageMaxReset();
#endif

#ifdef DEBUG
  Serial.print(F("Bank/Voice from EEPROM ["));
  Serial.print(configuration.bank, DEC);
  Serial.print(F("/"));
  Serial.print(configuration.voice, DEC);
  Serial.println(F("]"));
  show_patch();
#endif

  Serial.print(F("AUDIO_BLOCK_SAMPLES="));
  Serial.print(AUDIO_BLOCK_SAMPLES);
  Serial.print(F(" (Time per block="));
  Serial.print(1000000 / (SAMPLE_RATE / AUDIO_BLOCK_SAMPLES));
  Serial.println(F("ms)"));

#if defined (DEBUG) && defined (SHOW_CPU_LOAD_MSEC)
  show_cpu_and_mem_usage();
#endif

  AudioInterrupts();

  Serial.println(F("<setup end>"));
}

void loop()
{
#ifdef OLD
  int16_t* audio_buffer; // pointer to AUDIO_BLOCK_SAMPLES * int16_t
  const uint16_t audio_block_time_us = 1000000 / (SAMPLE_RATE / AUDIO_BLOCK_SAMPLES);
#endif

  //while (42 == 42)
  //{
#ifdef OLD
  // Main sound calculation
  if (queue1.available() && fill_audio_buffer > audio_block_time_us - 10)
  {
    fill_audio_buffer = 0;

    audio_buffer = queue1.getBuffer();

    elapsedMicros t1;
    for (uint8_t i = 0; i < NUM_DEXED; i++)
    {
      MicroDexed[i]->getSamples(AUDIO_BLOCK_SAMPLES, audio_buffer);
    }
    if (t1 > audio_block_time_us) // everything greater 2.9ms is a buffer underrun!
      xrun++;
    if (t1 > render_time_max)
      render_time_max = t1;
    if (peak1.available())
    {
      if (peak1.read() > 0.99)
        peak++;
    }
    queue1.playBuffer();
  }
#endif

  // EEPROM update handling
  if (autostore >= AUTOSTORE_MS && active_voices == 0 && eeprom_update_flag == true)
  {
    // only store configuration data to EEPROM when AUTOSTORE_MS is reached and no voices are activated anymore
    eeprom_update();
  }

  // MIDI input handling
  check_midi_devices();

  // CONTROL-RATE-EVENT-HANDLING
  if (control_rate > CONTROL_RATE_MS)
  {
#ifdef ENABLE_LCD_UI
    // LCD Menu
    LCDML.loop();

    // initial starts voice selection menu as default
    if (menu_state == MENU_START)
    {
      menu_state = MENU_VOICE;
      UI_func_voice_selection(0);
    }
#endif
    control_rate = 0;

    // check for value changes and unused voices
    soften_volume.tick();

    for (uint8_t i = 0; i < NUM_DEXED; i++)
    {
      active_voices = MicroDexed[i]->getNumNotesPlaying();

      soften_filter_res[i].tick();
      soften_filter_cut[i].tick();

      if (soften_filter_res[i].running())
      {
        // soften filter resonance value
        MicroDexed[i]->fx.Reso = soften_filter_res[i].value();

#ifdef DEBUG
        Serial.print(F("Filter-Resonance: "));
        Serial.print(MicroDexed[i]->fx.Reso, 5);
        Serial.print(F(" Filter-Resonance step: "));
        Serial.print(soften_filter_res[i].steps());
        Serial.print(F(" Filter-Resonance diff: "));
        Serial.println(soften_filter_res[i].diff(), 5);
#endif
      }
      // soften filter cutoff value
      if (soften_filter_cut[i].running())
      {
        MicroDexed[i]->fx.Cutoff = soften_filter_cut[i].value();
#ifdef DEBUG
        Serial.print(F("Filter-Cutoff: "));
        Serial.print(MicroDexed[i]->fx.Cutoff, 5);
        Serial.print(F(" Filter-Cutoff step: "));
        Serial.print(soften_filter_cut[i].steps());
        Serial.print(F(" Filter-Cutoff diff: "));
        Serial.println(soften_filter_cut[i].diff(), 5);
#endif
      }
    }
    if (soften_volume.running())
    {
      set_volume(soften_volume.value(), configuration.pan, configuration.mono);
#ifdef DEBUG
      Serial.print(F("Volume: "));
      Serial.print(configuration.vol, DEC);
      Serial.print(F(" step: "));
      Serial.print(soften_volume.steps());
      Serial.print(F(" diff: "));
      Serial.println(soften_volume.diff(), 5);
#endif
    }
  }

#if defined (DEBUG) && defined (SHOW_CPU_LOAD_MSEC)
  if (cpu_mem_millis >= SHOW_CPU_LOAD_MSEC)
  {
    if (peak1.available())
    {
      if (peak1.read() > 0.99)
        peak++;
    }
    cpu_mem_millis -= SHOW_CPU_LOAD_MSEC;
    show_cpu_and_mem_usage();
  }
#endif
  //}
}

/******************************************************************************
   MIDI MESSAGE HANDLER
 ******************************************************************************/
void handleNoteOn(byte inChannel, byte inNumber, byte inVelocity)
{
  if (checkMidiChannel(inChannel))
  {
    MicroDexed[0]->keydown(inNumber, inVelocity);
  }
}

void handleNoteOff(byte inChannel, byte inNumber, byte inVelocity)
{
  if (checkMidiChannel(inChannel))
  {
    MicroDexed[0]->keyup(inNumber);
  }
}

void handleControlChange(byte inChannel, byte inCtrl, byte inValue)
{
  if (checkMidiChannel(inChannel))
  {
#ifdef DEBUG
    Serial.print(F("CC#"));
    Serial.print(inCtrl, DEC);
    Serial.print(F(":"));
    Serial.println(inValue, DEC);
#endif

    switch (inCtrl) {
      case 0:
        if (inValue < MAX_BANKS)
        {
          configuration.bank = inValue;
        }
        break;
      case 1:
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->controllers.modwheel_cc = inValue;
          MicroDexed[i]->controllers.refresh();
        }
        break;
      case 2:
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->controllers.breath_cc = inValue;
          MicroDexed[i]->controllers.refresh();
        }
        break;
      case 4:
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->controllers.foot_cc = inValue;
          MicroDexed[i]->controllers.refresh();
        }
        break;
      case 7: // Volume
        configuration.vol = map(inValue, 0, 0x7f, VOLUME_MIN, VOLUME_MAX);
        soften_volume.update(configuration.vol, SOFTEN_VALUE_CHANGE_STEPS);
        break;
      case 10: // Pan
        configuration.pan = map(inValue, 0, 0x7f, PANORAMA_MIN, PANORAMA_MAX);
        set_volume(configuration.vol, configuration.pan, configuration.mono);
        break;
      case 32: // BankSelect LSB
        configuration.bank = inValue;
        break;
      case 64:
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->setSustain(inValue > 63);
          if (!MicroDexed[i]->getSustain()) {
            for (uint8_t note = 0; note < MicroDexed[0]->getMaxNotes(); note++) {
              if (MicroDexed[i]->voices[note].sustained && !MicroDexed[i]->voices[note].keydown) {
                MicroDexed[i]->voices[note].dx7_note->keyup();
                MicroDexed[i]->voices[note].sustained = false;
              }
            }
          }
        }
        break;
      case 103:  // CC 103: filter resonance
        configuration.filter_resonance = map(inValue, 0, 0x7f, FILTER_RESONANCE_MIN, FILTER_RESONANCE_MAX);
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->fx.Reso = configuration.filter_resonance / 100.0;
        }
        break;
      case 104:  // CC 104: filter cutoff
        configuration.filter_cutoff = map(inValue, 0, 0x7f, FILTER_CUTOFF_MIN, FILTER_CUTOFF_MAX);
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->fx.Cutoff = configuration.filter_cutoff / 100.0;
        }
        break;
      case 105:  // CC 105: delay time
        configuration.delay_time = map(inValue, 0, 0x7f, DELAY_TIME_MIN, DELAY_TIME_MAX);
        delay1.delay(0, configuration.delay_time * 10);
      case 106:  // CC 106: delay feedback
        configuration.delay_feedback = map(inValue, 0, 0x7f, DELAY_FEEDBACK_MIN , DELAY_FEEDBACK_MAX);
        delay_fb_mixer.gain(1, configuration.delay_feedback / 100.0 ); // amount of feedback
        delay_fb_mixer.gain(0, 1.0 - configuration.delay_feedback / 100.0); // original signalbreak;
        break;
      case 107:  // CC 107: delay volume
        configuration.delay_level = map(inValue, 0, 0x7f, DELAY_LEVEL_MIN, DELAY_LEVEL_MAX);
        master_mixer_r.gain(2, configuration.delay_level / 100.0);
        master_mixer_l.gain(2, configuration.delay_level / 100.0);
        break;
      case 120:
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->panic();
        }
        break;
      case 121:
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->resetControllers();
        }
        break;
      case 123:
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->notesOff();
        }
        break;
      case 126:
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->setMonoMode(true);
        }
        break;
      case 127:
        for (uint8_t i = 0; i < NUM_DEXED; i++)
        {
          MicroDexed[i]->setMonoMode(false);
        }
        break;
    }
  }
}

void handleAfterTouch(byte inChannel, byte inPressure)
{
  for (uint8_t i = 0; i < NUM_DEXED; i++)
  {
    MicroDexed[i]->controllers.aftertouch_cc = inPressure;
    MicroDexed[i]->controllers.refresh();
  }
}

void handlePitchBend(byte inChannel, int inPitch)
{
  for (uint8_t i = 0; i < NUM_DEXED; i++)
  {
    MicroDexed[i]->controllers.values_[kControllerPitch] = inPitch + 0x2000; // -8192 to +8191 --> 0 to 16383
  }
}

void handleProgramChange(byte inChannel, byte inProgram)
{
  if (inProgram < MAX_VOICES)
  {
#ifdef DISPLAY_LCD_SPI
    change_disp_sd(false);
#endif
    load_sysex(configuration.bank, inProgram);
#ifdef DISPLAY_LCD_SPI
    change_disp_sd(true);
#endif
  }
}

void handleSystemExclusive(byte * sysex, uint len)
{
  /*
    SYSEX MESSAGE: Parameter Change
    -------------------------------
       bits    hex  description

     11110000  F0   Status byte - start sysex
     0iiiiiii  43   ID # (i=67; Yamaha)
     0sssnnnn  10   Sub-status (s=1) & channel number (n=0; ch 1)
     0gggggpp  **   parameter group # (g=0; voice, g=2; function)
     0ppppppp  **   parameter # (these are listed in next section)
                     Note that voice parameter #'s can go over 128 so
                     the pp bits in the group byte are either 00 for
                     par# 0-127 or 01 for par# 128-155. In the latter case
                     you add 128 to the 0ppppppp byte to compute par#.
     0ddddddd  **   data byte
     11110111  F7   Status - end sysex
  */

#ifdef DEBUG
  Serial.print(F("SYSEX-Data["));
  Serial.print(len, DEC);
  Serial.print(F("]"));
  for (uint8_t i = 0; i < len; i++)
  {
    Serial.print(F(" "));
    Serial.print(sysex[i], DEC);
  }
  Serial.println();
#endif

  if (!checkMidiChannel((sysex[2] & 0x0f) + 1 ))
  {
#ifdef DEBUG
    Serial.println(F("SYSEX-MIDI-Channel mismatch"));
#endif
    return;
  }

  if (sysex[1] != 0x43) // check for Yamaha sysex
  {
#ifdef DEBUG
    Serial.println(F("E: SysEx vendor not Yamaha."));
#endif
    return;
  }

#ifdef DEBUG
  Serial.print(F("Substatus: ["));
  Serial.print((sysex[2] & 0x70) >> 4);
  Serial.println(F("]"));
#endif

  // parse parameter change
  if (len == 7)
  {
    if (((sysex[3] & 0x7c) >> 2) != 0 && ((sysex[3] & 0x7c) >> 2) != 2)
    {
#ifdef DEBUG
      Serial.println(F("E: Not a SysEx parameter or function parameter change."));
#endif
      return;
    }
    if (sysex[6] != 0xf7)
    {
#ifdef DEBUG
      Serial.println(F("E: SysEx end status byte not detected."));
#endif
      return;
    }

    sysex[4] &= 0x7f;
    sysex[5] &= 0x7f;

    uint8_t data_index;

    if (((sysex[3] & 0x7c) >> 2) == 0)
    {
      MicroDexed[0]->notesOff();
      MicroDexed[0]->data[sysex[4] + ((sysex[3] & 0x03) * 128)] = sysex[5]; // set parameter
      MicroDexed[0]->doRefreshVoice();
      data_index = sysex[4] + ((sysex[3] & 0x03) * 128);
    }
    else
    {
      MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET - 63 + sysex[4]] = sysex[5]; // set function parameter
      MicroDexed[0]->controllers.values_[kControllerPitchRange] = MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_PITCHBEND_RANGE];
      MicroDexed[0]->controllers.values_[kControllerPitchStep] = MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_PITCHBEND_STEP];
      MicroDexed[0]->controllers.wheel.setRange(MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_MODWHEEL_RANGE]);
      MicroDexed[0]->controllers.wheel.setTarget(MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_MODWHEEL_ASSIGN]);
      MicroDexed[0]->controllers.foot.setRange(MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_FOOTCTRL_RANGE]);
      MicroDexed[0]->controllers.foot.setTarget(MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_FOOTCTRL_ASSIGN]);
      MicroDexed[0]->controllers.breath.setRange(MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_BREATHCTRL_RANGE]);
      MicroDexed[0]->controllers.breath.setTarget(MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_BREATHCTRL_ASSIGN]);
      MicroDexed[0]->controllers.at.setRange(MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_AT_RANGE]);
      MicroDexed[0]->controllers.at.setTarget(MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_AT_ASSIGN]);
      MicroDexed[0]->controllers.masterTune = (MicroDexed[0]->data[DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_MASTER_TUNE] * 0x4000 << 11) * (1.0 / 12);
      MicroDexed[0]->controllers.refresh();
      data_index = DEXED_GLOBAL_PARAMETER_OFFSET - 63 + sysex[4];
    }
#ifdef DEBUG
    Serial.print(F("SysEx"));
    if (((sysex[3] & 0x7c) >> 2) == 0)
      Serial.print(F(" function"));
    Serial.print(F(" parameter "));
    Serial.print(sysex[4], DEC);
    Serial.print(F(" = "));
    Serial.print(sysex[5], DEC);
    Serial.print(F(", data_index = "));
    Serial.println(data_index, DEC);
#endif
  }
  else if (len == 163)
  {
    int32_t bulk_checksum_calc = 0;
    int8_t bulk_checksum = sysex[161];

    // 1 Voice bulk upload
#ifdef DEBUG
    Serial.println(F("One Voice bulk upload"));
#endif

    if (sysex[162] != 0xf7)
    {
#ifdef DEBUG
      Serial.println(F("E: Found no SysEx end marker."));
#endif
      return;
    }

    if ((sysex[3] & 0x7f) != 0)
    {
#ifdef DEBUG
      Serial.println(F("E: Not a SysEx voice bulk upload."));
#endif
      return;
    }

    if (((sysex[4] << 7) | sysex[5]) != 0x9b)
    {
#ifdef DEBUG
      Serial.println(F("E: Wrong length for SysEx voice bulk upload (not 155)."));
#endif
      return;
    }

    // checksum calculation
    for (uint8_t i = 0; i < 155 ; i++)
    {
      bulk_checksum_calc -= sysex[i + 6];
    }
    bulk_checksum_calc &= 0x7f;

    if (bulk_checksum_calc != bulk_checksum)
    {
#ifdef DEBUG
      Serial.print(F("E: Checksum error for one voice [0x"));
      Serial.print(bulk_checksum, HEX);
      Serial.print(F("/0x"));
      Serial.print(bulk_checksum_calc, HEX);
      Serial.println(F("]"));
#endif
      return;
    }

    // load sysex-data into voice memory
    MicroDexed[0]->loadVoiceParameters(&sysex[6]);
    //MicroDexed[0]->initGlobalParameters();

    // manipulate UI names and numbers
    strncpy(voice_name, (char *)&sysex[151], sizeof(voice_name) - 1);
    Serial.print(F("Got voice ["));
    Serial.print(voice_name);
    Serial.println(F("]."));
  }
#ifdef DEBUG
  else
    Serial.println(F("E: SysEx parameter length wrong."));
#endif
}

void handleTimeCodeQuarterFrame(byte data)
{
  ;
}

void handleAfterTouchPoly(byte inChannel, byte inNumber, byte inVelocity)
{
  ;
}

void handleSongSelect(byte inSong)
{
  ;
}

void handleTuneRequest(void)
{
  ;
}

void handleClock(void)
{
  midi_timing_counter++;
  if (midi_timing_counter % 24 == 0)
  {
    midi_timing_quarter = midi_timing_timestep;
    midi_timing_counter = 0;
    midi_timing_timestep = 0;
    // Adjust delay control here
#ifdef DEBUG
    Serial.print(F("MIDI Clock: "));
    Serial.print(60000 / midi_timing_quarter, DEC);
    Serial.print(F("bpm ("));
    Serial.print(midi_timing_quarter, DEC);
    Serial.println(F("ms per quarter)"));
#endif
  }
}

void handleStart(void)
{
  ;
}

void handleContinue(void)
{
  ;
}

void handleStop(void)
{
  ;
}

void handleActiveSensing(void)
{
  ;
}

void handleSystemReset(void)
{
#ifdef DEBUG
  Serial.println(F("MIDI SYSEX RESET"));
#endif
  MicroDexed[0]->notesOff();
  MicroDexed[0]->panic();
  MicroDexed[0]->resetControllers();
}

/******************************************************************************
   MIDI HELPER
 ******************************************************************************/
bool checkMidiChannel(byte inChannel)
{
  // check for MIDI channel
  if (configuration.midi_channel == MIDI_CHANNEL_OMNI)
  {
    return (true);
  }
  else if (inChannel != configuration.midi_channel)
  {
#ifdef DEBUG
    Serial.print(F("Ignoring MIDI data on channel "));
    Serial.print(inChannel);
    Serial.print(F("(listening on "));
    Serial.print(configuration.midi_channel);
    Serial.println(F(")"));
#endif
    return (false);
  }
  return (true);
}

/******************************************************************************
   VOLUME HELPER
 ******************************************************************************/

void set_volume(uint8_t v, int8_t p, uint8_t m)
{
  configuration.vol = v;
  configuration.pan = p;
  configuration.mono = m;

  uint16_t tmp = v / 100.0 * 1023.0 + 0.5;
  float tmp2 = mapfloat(configuration.pan, PANORAMA_MIN, PANORAMA_MAX, 0.0, 1.0);
  float tmp3 = (float)(tmp * (tmp + 2)) / (float)(1 << 20);

#ifdef DEBUG
  Serial.print(F("Setting volume: VOL="));
  Serial.print(v, DEC);
  Serial.print(F("["));
  Serial.print(tmp3, 3);
  Serial.print(F("] PAN="));
  Serial.print(p, DEC);
  Serial.print(F("["));
  Serial.print(tmp2, 3);
  Serial.print(F("] "));
  Serial.print(tmp3 * sinf(tmp2 * PI / 2), 3);
  Serial.print(F("/"));
  Serial.println(tmp3 * cosf(tmp2 * PI / 2), 3);
#endif

  // float v = (float)(a * (a + 2))/(float)(1 << 20); //  (pseudo-) logarithmic curve for volume control
  // http://files.csound-tutorial.net/floss_manual/Release03/Cs_FM_03_ScrapBook/b-panning-and-spatialization.html
  volume_r.gain(tmp3 * sinf(tmp2 * PI / 2));
  volume_l.gain(tmp3 * cosf(tmp2 * PI / 2));

  switch (m)
  {
    case 0: // stereo
      stereomono1.stereo(true);
      modchorus_inverter.gain(-1.0); // stereo mode
      break;
    case 1: // mono both
      stereomono1.stereo(false);
      configuration.pan = 0.5;
      modchorus_inverter.gain(1.0); // stereo mode
      break;
    case 2: // mono right
      volume_l.gain(0.0);
      stereomono1.stereo(false);
      configuration.pan = 0.5;
      modchorus_inverter.gain(1.0); // stereo mode
      break;
    case 3: // mono left
      volume_r.gain(0.0);
      stereomono1.stereo(false);
      configuration.pan = 0.5;
      modchorus_inverter.gain(1.0); // stereo mode
      break;
  }
}

// https://www.dr-lex.be/info-stuff/volumecontrols.html#table1
inline float logvol(float x)
{
  return (0.001 * expf(6.908 * x));
}

/******************************************************************************
   EEPROM HELPER
 ******************************************************************************/

void initial_values_from_eeprom(void)
{
  uint32_t checksum;
  config_t tmp_conf;

  EEPROM_readAnything(EEPROM_START_ADDRESS, tmp_conf);
  checksum = crc32((byte*)&tmp_conf + 4, sizeof(tmp_conf) - 4);

#ifdef DEBUG
  Serial.print(F("EEPROM checksum: 0x"));
  Serial.print(tmp_conf.checksum, HEX);
  Serial.print(F(" / 0x"));
  Serial.print(checksum, HEX);
#endif

  if (checksum != tmp_conf.checksum)
  {
#ifdef DEBUG
    Serial.print(F(" - mismatch -> initializing EEPROM!"));
#endif
    configuration.checksum = 0xffff;
    configuration.bank = SYSEXBANK_DEFAULT;
    configuration.voice = SYSEXSOUND_DEFAULT;
    configuration.vol = VOLUME_DEFAULT;
    configuration.pan = PANORAMA_DEFAULT;
    configuration.mono = MONO_DEFAULT;
    configuration.midi_channel = DEFAULT_MIDI_CHANNEL;
    configuration.reverb_roomsize = REVERB_ROOMSIZE_DEFAULT;
    configuration.reverb_damping = REVERB_DAMPING_DEFAULT;
    configuration.reverb_level = REVERB_LEVEL_DEFAULT;
    configuration.chorus_frequency = CHORUS_FREQUENCY_DEFAULT;
    configuration.chorus_waveform = CHORUS_WAVEFORM_DEFAULT;
    configuration.chorus_depth = CHORUS_DEPTH_DEFAULT;
    configuration.chorus_level = CHORUS_LEVEL_DEFAULT;
    configuration.delay_time = DELAY_TIME_DEFAULT;
    configuration.delay_feedback = DELAY_FEEDBACK_DEFAULT;
    configuration.delay_level = DELAY_LEVEL_DEFAULT;
    configuration.filter_cutoff = FILTER_CUTOFF_DEFAULT;
    configuration.filter_resonance = FILTER_RESONANCE_DEFAULT;
    configuration.loudness = LOUDNESS_DEFAULT;
    configuration.polyphony = POLYPHONY_DEFAULT;
    configuration.engine = ENGINE_DEFAULT;
    eeprom_update();
  }
  else
  {
    EEPROM_readAnything(EEPROM_START_ADDRESS, configuration);
    Serial.print(F(" - OK, loading!"));
  }
#ifdef DEBUG
  Serial.println();
  show_configuration();
#endif

  if (configuration.vol > 100)
    configuration.vol = 100;
}

void eeprom_write(void)
{
  autostore = 0;
  eeprom_update_flag = true;
}

void eeprom_update(void)
{
  eeprom_update_flag = false;
  configuration.checksum = crc32((byte*)&configuration + 4, sizeof(configuration) - 4);
  EEPROM_writeAnything(EEPROM_START_ADDRESS, configuration);
  Serial.println(F("Updating EEPROM with configuration data"));
}

uint32_t crc32(byte * calc_start, uint16_t calc_bytes) // base code from https://www.arduino.cc/en/Tutorial/EEPROMCrc
{
  const uint32_t crc_table[16] =
  {
    0x00000000, 0x1db71064, 0x3b6e20c8, 0x26d930ac,
    0x76dc4190, 0x6b6b51f4, 0x4db26158, 0x5005713c,
    0xedb88320, 0xf00f9344, 0xd6d6a3e8, 0xcb61b38c,
    0x9b64c2b0, 0x86d3d2d4, 0xa00ae278, 0xbdbdf21c
  };
  uint32_t crc = ~0L;

  for (byte* index = calc_start ; index < (calc_start + calc_bytes) ; ++index)
  {
    crc = crc_table[(crc ^ *index) & 0x0f] ^ (crc >> 4);
    crc = crc_table[(crc ^ (*index >> 4)) & 0x0f] ^ (crc >> 4);
    crc = ~crc;
  }

  return (crc);
}

/******************************************************************************
   DEBUG HELPER
 ******************************************************************************/

#if defined (DEBUG) && defined (SHOW_CPU_LOAD_MSEC)
void show_cpu_and_mem_usage(void)
{
  Serial.print(F("CPU: "));
  Serial.print(AudioProcessorUsage(), 2);
  Serial.print(F("%   CPU MAX: "));
  Serial.print(AudioProcessorUsageMax(), 2);
  Serial.print(F("%  MEM: "));
  Serial.print(AudioMemoryUsage(), DEC);
  Serial.print(F("   MEM MAX: "));
  Serial.print(AudioMemoryUsageMax(), DEC);
  Serial.print(F("   RENDER_TIME_MAX: "));
  Serial.print(MicroDexed[0]->render_time_max, DEC);
  Serial.print(F("   XRUN: "));
  Serial.print(MicroDexed[0]->xrun, DEC);
  Serial.print(F("   OVERLOAD: "));
  Serial.print(MicroDexed[0]->overload, DEC);
  Serial.print(F("   PEAK: "));
  Serial.print(peak, DEC);
  Serial.print(F(" BLOCKSIZE: "));
  Serial.print(AUDIO_BLOCK_SAMPLES, DEC);
  Serial.print(F(" ACTIVE_VOICES: "));
  Serial.print(active_voices, DEC);
  Serial.println();
  AudioProcessorUsageMaxReset();
  AudioMemoryUsageMaxReset();
  MicroDexed[0]->render_time_max = 0;
}
#endif

#ifdef DEBUG
void show_configuration(void)
{
  Serial.println();
  Serial.println(F("CONFIGURATION:"));
  Serial.print(F("Checksum         0x")); Serial.println(configuration.checksum, HEX);
  Serial.print(F("Bank             ")); Serial.println(configuration.bank, DEC);
  Serial.print(F("Voice            ")); Serial.println(configuration.voice, DEC);
  Serial.print(F("Volume           ")); Serial.println(configuration.vol, DEC);
  Serial.print(F("Panorama         ")); Serial.println(configuration.pan, DEC);
  Serial.print(F("Mono             ")); Serial.println(configuration.mono, DEC);
  Serial.print(F("MIDI-Channel     ")); Serial.println(configuration.midi_channel, DEC);
  Serial.print(F("Reverb Roomsize  ")); Serial.println(configuration.reverb_roomsize, DEC);
  Serial.print(F("Reverb Damping   ")); Serial.println(configuration.reverb_damping, DEC);
  Serial.print(F("Reverb Level     ")); Serial.println(configuration.reverb_level, DEC);
  Serial.print(F("Chorus Frequency ")); Serial.println(configuration.chorus_frequency, DEC);
  Serial.print(F("Chorus Waveform  ")); Serial.println(configuration.chorus_waveform, DEC);
  Serial.print(F("Chorus Depth     ")); Serial.println(configuration.chorus_depth, DEC);
  Serial.print(F("Chorus Level     ")); Serial.println(configuration.chorus_level, DEC);
  Serial.print(F("Delay Time       ")); Serial.println(configuration.delay_time, DEC);
  Serial.print(F("Delay Feedback   ")); Serial.println(configuration.delay_feedback, DEC);
  Serial.print(F("Delay Level      ")); Serial.println(configuration.delay_level, DEC);
  Serial.print(F("Filter Cutoff    ")); Serial.println(configuration.filter_cutoff, DEC);
  Serial.print(F("Filter Resonance ")); Serial.println(configuration.filter_resonance, DEC);
  Serial.print(F("Loudness         ")); Serial.println(configuration.loudness, DEC);
  Serial.print(F("Polyphony        ")); Serial.println(configuration.polyphony, DEC);
  Serial.print(F("Engine           ")); Serial.println(configuration.engine, DEC);
  Serial.println();
}

void show_patch(void)
{
  uint8_t i;
  char voicename[VOICE_NAME_LEN];

  memset(voicename, 0, sizeof(voicename));
  for (i = 0; i < 6; i++)
  {
    Serial.print(F("OP"));
    Serial.print(6 - i, DEC);
    Serial.println(F(": "));
    Serial.println(F("R1 | R2 | R3 | R4 | L1 | L2 | L3 | L4 LEV_SCL_BRK_PT | SCL_LEFT_DEPTH | SCL_RGHT_DEPTH"));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_EG_R1], DEC);
    Serial.print(F(" "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_EG_R2], DEC);
    Serial.print(F(" "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_EG_R3], DEC);
    Serial.print(F(" "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_EG_R4], DEC);
    Serial.print(F(" "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_EG_L1], DEC);
    Serial.print(F(" "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_EG_L2], DEC);
    Serial.print(F(" "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_EG_L3], DEC);
    Serial.print(F(" "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_EG_L4], DEC);
    Serial.print(F("        "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_LEV_SCL_BRK_PT], DEC);
    Serial.print(F("             "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_SCL_LEFT_DEPTH], DEC);
    Serial.print(F("             "));
    Serial.println(MicroDexed[0]->data[(i * 21) + DEXED_OP_SCL_RGHT_DEPTH], DEC);
    Serial.println(F("SCL_L_CURVE | SCL_R_CURVE | RT_SCALE | AMS | KVS | OUT_LEV | OP_MOD | FRQ_C | FRQ_F | DETUNE"));
    Serial.print(F("      "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_SCL_LEFT_CURVE], DEC);
    Serial.print(F("         "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_SCL_RGHT_CURVE], DEC);
    Serial.print(F("         "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_OSC_RATE_SCALE], DEC);
    Serial.print(F("        "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_AMP_MOD_SENS], DEC);
    Serial.print(F("     "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_KEY_VEL_SENS], DEC);
    Serial.print(F("      "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_OUTPUT_LEV], DEC);
    Serial.print(F("      "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_OSC_MODE], DEC);
    Serial.print(F("    "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_FREQ_COARSE], DEC);
    Serial.print(F("     "));
    Serial.print(MicroDexed[0]->data[(i * 21) + DEXED_OP_FREQ_FINE], DEC);
    Serial.print(F("     "));
    Serial.println(MicroDexed[0]->data[(i * 21) + DEXED_OP_OSC_DETUNE], DEC);
  }
  Serial.println(F("PR1 | PR2 | PR3 | PR4 | PL1 | PL2 | PL3 | PL4"));
  Serial.print(F(" "));
  for (i = 0; i < 8; i++)
  {
    Serial.print(MicroDexed[0]->data[DEXED_VOICE_OFFSET + i], DEC);
    Serial.print(F("  "));
  }
  Serial.println();
  Serial.print(F("ALG: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_ALGORITHM], DEC);
  Serial.print(F("FB: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_FEEDBACK], DEC);
  Serial.print(F("OKS: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_OSC_KEY_SYNC], DEC);
  Serial.print(F("LFO SPD: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_LFO_SPEED], DEC);
  Serial.print(F("LFO_DLY: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_LFO_DELAY], DEC);
  Serial.print(F("LFO PMD: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_LFO_PITCH_MOD_DEP], DEC);
  Serial.print(F("LFO_AMD: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_LFO_AMP_MOD_DEP], DEC);
  Serial.print(F("LFO_SYNC: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_LFO_SYNC], DEC);
  Serial.print(F("LFO_WAVEFRM: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_LFO_WAVE], DEC);
  Serial.print(F("LFO_PMS: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_LFO_PITCH_MOD_SENS], DEC);
  Serial.print(F("TRNSPSE: "));
  Serial.println(MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_TRANSPOSE], DEC);
  Serial.print(F("NAME: "));
  strncpy(voicename, (char *)&MicroDexed[0]->data[DEXED_VOICE_OFFSET + DEXED_NAME], sizeof(voicename) - 1);
  Serial.print(F("["));
  Serial.print(voicename);
  Serial.println(F("]"));
  for (i = DEXED_GLOBAL_PARAMETER_OFFSET; i <= DEXED_GLOBAL_PARAMETER_OFFSET + DEXED_MAX_NOTES; i++)
  {
    Serial.print(i, DEC);
    Serial.print(F(": "));
    Serial.println(MicroDexed[0]->data[i]);
  }

  Serial.println();
}
#endif
